import os
import re
import sys
import logging
from typing import List

import django
from django.db import transaction
from django.contrib.gis.geos import GEOSGeometry, Polygon

from .exceptions import RegistrationError
from .context import Context
from .source import Source, S3Source, SwiftSource, LocalSource


logger = logging.getLogger(__name__)


class RegistrationResult:
    pass


class Backend:
    def register_item(self, item: Context) -> RegistrationResult:
        raise NotImplementedError


class EOxServerBackend(Backend):
    def __init__(self, instance_base_path: str, instance_name: str, mapping: dict, simplify_footprint_tolerance: int=None):
        self.mapping = mapping
        self.simplify_footprint_tolerance = simplify_footprint_tolerance
        path = os.path.join(instance_base_path, instance_name)
        if path not in sys.path:
            sys.path.append(path)

        os.environ.setdefault("DJANGO_SETTINGS_MODULE", f"{instance_name}.settings") # TODO: from config
        django.setup()

    def exists(self, source: Source, item: Context):
        from eoxserver.resources.coverages import models
        return models.Product.objects.filter(identifier=item.identifier).exists()

    def _get_storage_from_source(self, source: Source, path: str) -> list:
        return [source.name] if source.name else []

    @transaction.atomic
    def register(self, source: Source, item: Context, replace: bool) -> RegistrationResult:
        # ugly, ugly hack
        from eoxserver.resources.coverages import models
        from eoxserver.resources.coverages.registration.product import ProductRegistrator
        from eoxserver.resources.coverages.registration.browse import BrowseRegistrator
        from eoxserver.resources.coverages.registration.mask import MaskRegistrator
        from eoxserver.resources.coverages.registration.registrators.gdal import GDALRegistrator

        # get the mapping for this particular item
        mapping = self.mapping[item.product_type][item.product_level]
        metadata_file = item.metadata_files[0]

        storage = self._get_storage_from_source(source, item.path)

        try:
            models.ProductType.objects.get(name=mapping['product_type_name'])
        except models.ProductType.DoesNotExist:
            pass

        footprint = GEOSGeometry(item.metadata.pop('footprint'))

        product, _ = ProductRegistrator().register(
            metadata_locations=[storage + [metadata_file]],
            type_name=mapping['product_type_name'],
            replace=replace,
            extended_metadata=True,
            mask_locations=None,
            package_path=None,
            simplify_footprint_tolerance=self.simplify_footprint_tolerance,
            overrides=dict(
                identifier=item.identifier,
                footprint=footprint,
                **item.metadata
            ),
        )
        if not product.footprint or product.footprint.empty:
            raise RegistrationError("No footprint was extracted. full product: %s" % product)

        # insert the product in the to be associated collections
        for collection_id in mapping.get('collections', []):
            collection = models.Collection.objects.get(
                identifier=collection_id,
            )
            models.collection_insert_eo_object(collection, product)

        # register coverages and link them to the product
        for raster_identifier, coverage_type_name in mapping.get('coverages', {}).items():
            raster_item = item.raster_files.get(raster_identifier)

            report = GDALRegistrator().register(
                data_locations=[storage + [raster_item]],
                metadata_locations=[storage + [metadata_file]],
                coverage_type_name=coverage_type_name,
                overrides={
                    "identifier": f'{product.identifier}__{raster_identifier}__coverage',
                    "footprint": None,
                },
                replace=replace,
            )
            logger.debug("Adding coverage to product")
            models.product_add_coverage(product, report.coverage)

        # register browses
        for raster_identifier, browse_type_name in mapping.get('browses', {}).items():
            raster_item = item.raster_files.get(raster_identifier)

            raster_item = '/'.join(raster_item.split('/')[1:])
            logger.info(f"Adding browse {browse_type_name or 'default'} {raster_item} to product")

            BrowseRegistrator().register(
                product.identifier,
                storage + [raster_item],
                browse_type_name,
            )

        # register masks
        for mask_identifier, mask_type_name in mapping.get('masks', {}).items():
            mask_item = item.mask_files.get(mask_identifier)
            if mask_item:
                logger.info(f"Adding mask {mask_type_name} to product")
                MaskRegistrator().register(
                    product.identifier,
                    storage + [mask_item],
                    mask_type_name,
                )


BACKENDS = {
    'eoxserver': EOxServerBackend
}

def get_backends(config: dict, path: str) -> List[Backend]:
    cfg_backends = config['backends']

    backends = [
        BACKENDS[cfg_backend['type']](
            *cfg_backend.get('args', []),
            **cfg_backend.get('kwargs', {}),
        )
        for cfg_backend in cfg_backends
        if not cfg_backend.get('filter') or re.match(cfg_backend['filter'], path)
    ]

    if not backends:
        raise RegistrationError(f'Could not find a suitable backend for the path {path}')

    return backends

